#include "Lexer.h"

#include <stdlib.h> //exit(1)
#include <fstream>


Lexer::Lexer(vector<Regexp*> lexical_description, set<char> alphabet)
{
  //
  // Build the lexer regexp E = (E_1.#_1 U ... U E_n.#_n)*
  //

  Regexp* lexer_regexp;
  set<char> markers;

  int first_iteration = true;
  for(int i=0; i<lexical_description.size(); i++)
    //For each token's regexp...
    {
      Regexp* E_i = lexical_description[i];
      Regexp* marked_regexp = new Regexp(CONCAT,E_i,new Regexp((char)(i+1)));
      //Add a marker to 'regexp'
      
      if(first_iteration)
	lexer_regexp = marked_regexp;
      else
	lexer_regexp = new Regexp(UNION,lexer_regexp,marked_regexp);
      //union
      
      first_iteration = false;
      markers.insert((char)(i+1));
    }
    //E_1.#_1 U ... U E_n.#_n

  lexer_regexp = new Regexp(STAR,lexer_regexp);
  //E = (E_1.#_1 U ... U E_n.#_n)*

  marker_alphabet = markers;

  //
  // Build the corresponding automaton
  //

  set<char> sigma = alphabet;
  for(set<char>::const_iterator it = markers.begin(); it != markers.end(); ++it)
    sigma.insert(*it);
  //sigma = alphabet U markers

  Automaton* thompson = new Automaton(sigma,lexer_regexp);
  //1) Thompson construction

  Automaton* det = thompson->determinize();
  det->complete();
  //2) Determinization + completion

  a = det->minimize();
  //3) Minimization

  ofstream output_file;
  output_file.open("automata.dot");
  thompson->print_dot(output_file);
  det->print_dot(output_file);
  a->print_dot(output_file);
  output_file.close();
}


void
Lexer::print_code(ostream& sout)
{
  pair<pair< vector<vector<int> > , map<char,int> >, vector<int> > tt = 
    a->transition_table(marker_alphabet);
  vector<vector<int> > table = tt.first.first;
  map<char,int> letter_rank = tt.first.second;
  vector<int> state_token = tt.second;
  //Get :
  //  transition table + 
  //  mapping letter -> rank + 
  //  mapping state -> token

  sout << "#include <string.h>" << endl << endl;

  //
  // Emit the transition table
  //

  sout << "int next[" << a->state.size() << "][" << letter_rank.size() << "] ="
       << " {"
       << endl;

  for(int i=0; i<table.size(); i++)
    {
      sout << "  {";
      for(int j=0; j<table[i].size(); j++)
	{
	  sout << table[i][j];
	  if(j != table[i].size()-1) 
	    sout << ", ";
	}
      sout << "}";
      if(i != table.size()-1)
	sout << ",";
      sout << endl;
    }
  sout << "};" << endl << endl;

  //
  // Emit the state -> token mapping
  //

  int min_char = state_token[0];
  for(int i=1; i<state_token.size(); i++)
    if(state_token[i]<min_char)
      min_char = state_token[i]; //min char#

  sout << "int token[" << a->state.size() << "] ="
       << " {";
  for(int i=0; i<a->state.size(); i++)
    {
      sout << (int)(state_token[i]) - min_char;
      if(i != a->state.size()-1) 
	sout << ", ";
    }
  sout << "};" << endl << endl;

  //
  // Emit external user-defined functions
  //

  sout << "extern void reset_input_flow();" << endl
       << "extern char read_next();" << endl
       << "extern void unread();" << endl
       << "extern char Eof;" << endl
       << "extern void accept(int token, char* string);" << endl
       << "extern void error(int car_number);" << endl << endl;

  //
  // Emit the letter mapping
  //
  sout << "int char_number;" << endl << endl;

  sout << "int letter_rank(char c)" << endl
       << "{" << endl;
  for(map<char,int>::const_iterator it = letter_rank.begin(); 
      it != letter_rank.end(); ++it)
    sout << "  if(c == '" << it->first << "') return " << it->second << ";" 
	 << endl;
  sout << "  error(char_number);" << endl
       << "}" << endl << endl;

  sout << "char* strchar(char c)" << endl
       << "{" << endl
       << "  char* result = strdup(\"*\");" << endl
       << "  result[0] = c;" << endl
       << "  return result;" << endl
       << "}" << endl << endl;

  //
  // Then, emit the automaton
  //

  sout << "int yylex()" << endl
       << "{" << endl
       << "  reset_input_flow();" << endl
       << "  int state = " << a->state_rank(a->initial_state) << ";" << endl
       << "  char c = read_next();" << endl
       << "  char_number = 1;" << endl
       << "  char* yytext = strdup(\"\");" << endl
       << "  while(c != Eof)" << endl
       << "    {" << endl
       << "      int target_state = next[state][letter_rank(c)];" << endl
       << "      if(target_state == " << a->state_rank(a->default_state) << ")"
       << endl
       << "        {" << endl
       << "          if(token[state] == 0)" << endl
       << "            {" << endl
       << "              error(char_number);" << endl
       << "              return 1;" << endl
       << "            }" << endl
       << "          accept(token[state],yytext);" << endl
       << "          unread();" << endl
       << "          yytext = strchar(c);" << endl
       << "          target_state = " << a->state_rank(a->initial_state) << ";"
       << endl
       << "        }" << endl
       << "      else" << endl
       << "        yytext = strcat(yytext,strchar(c));" << endl
       << "      state = target_state;" << endl
       << "      c = read_next();" << endl
       << "      char_number++;" << endl
       << "    }" << endl
       << "    if(token[state] == 0)" << endl
       << "      {" << endl
       << "        error(char_number);" << endl
       << "        return 1;" << endl
       << "      }" << endl
       << "    accept(token[state],yytext);" << endl
       << "  return 0;" << endl
       << "}" << endl;
  
}
